"""Tests for flexmock stubbing feature."""

# pylint: disable=missing-docstring,no-member
from flexmock import flexmock
from flexmock._api import flexmock_teardown
from tests import some_module


class StubbingTestCase:
    def test_use_replace_with_for_callable_shortcut_kwargs(self):
        class Foo:
            def method(self):
                return "bar"

        instance = Foo()
        flexmock(instance, method=lambda: "baz")
        assert instance.method() == "baz"

    def test_should_replace_attributes_that_are_instances_of_classes(self):
        class Class1:
            pass

        class Class2:
            class1 = Class1()

        class2 = Class2()
        flexmock(class2, class1="test")
        assert class2.class1 == "test"

    def test_replace_non_callable_instance_attributes(self):
        class FooClass:
            def __init__(self):
                self.attribute = 1

        instance1 = FooClass()
        instance2 = FooClass()
        flexmock(instance1, attribute=2)
        flexmock(instance2, attribute=1)
        assert instance1.attribute == 2
        flexmock_teardown()
        assert instance1.attribute == 1

    def test_replace_non_callable_module_attributes(self):
        flexmock(some_module, MODULE_LEVEL_ATTRIBUTE="yay")
        assert some_module.MODULE_LEVEL_ATTRIBUTE == "yay"
        flexmock_teardown()
        assert some_module.MODULE_LEVEL_ATTRIBUTE == "test"

    def test_replace_non_callable_class_attributes(self):
        class FooClass:
            attribute = 1

        instance1 = FooClass()
        instance2 = FooClass()
        flexmock(instance1, attribute=2)
        assert instance1.attribute == 2
        assert instance2.attribute == 1
        flexmock_teardown()
        assert instance1.attribute == 1

    def test_fake_object_takes_properties(self):
        fake1 = flexmock(bar=property(lambda self: "baz"))
        fake2 = flexmock(foo=property(lambda self: "baz"))
        assert fake1.bar == "baz"
        assert fake2.foo == "baz"



"""Tests for flexmock teardown."""

# pylint: disable=missing-docstring,no-member
from flexmock import exceptions, flexmock
from flexmock._api import UPDATED_ATTRS, flexmock_teardown
from tests import some_module
from tests.utils import assert_raises


class TeardownTestCase:
    def test_flexmock_should_properly_restore_static_methods(self):
        class User:
            @staticmethod
            def get_stuff():
                return "ok!"

        assert isinstance(User.__dict__["get_stuff"], staticmethod)
        assert User.get_stuff() == "ok!"
        flexmock(User).should_receive("get_stuff")
        assert User.get_stuff() is None
        flexmock_teardown()
        assert User.get_stuff() == "ok!"
        assert isinstance(User.__dict__["get_stuff"], staticmethod)

    def test_flexmock_should_properly_restore_undecorated_static_methods(self):
        class User:
            def get_stuff():
                return "ok!"

            get_stuff = staticmethod(get_stuff)  # pylint: disable=no-staticmethod-decorator

        assert User.get_stuff() == "ok!"
        flexmock(User).should_receive("get_stuff")
        assert User.get_stuff() is None
        flexmock_teardown()
        assert User.get_stuff() == "ok!"

    def test_flexmock_should_properly_restore_module_level_functions(self):
        flexmock(some_module).should_receive("module_function").with_args(1, 2)
        assert some_module.module_function(1, 2) is None
        flexmock_teardown()
        assert some_module.module_function(1, 2) == -1

    def test_flexmock_should_revert_new_instances_on_teardown(self):
        class User:
            pass

        class Group:
            pass

        user = User()
        group = Group()
        flexmock(Group).new_instances(user)
        assert user is Group()
        flexmock_teardown()
        assert group.__class__ is Group().__class__

    def test_flexmock_should_cleanup_added_methods_and_attributes(self):
        class Group:
            pass

        group = Group()
        flexmock(Group)
        assert "should_receive" in Group.__dict__
        assert "should_receive" not in group.__dict__
        flexmock(group)
        assert "should_receive" in group.__dict__
        flexmock_teardown()
        for method in UPDATED_ATTRS:
            assert method not in Group.__dict__
            assert method not in group.__dict__

    def test_class_attributes_are_unchanged_after_mocking(self):
        class Base:
            @classmethod
            def class_method(cls):
                pass

            @staticmethod
            def static_method():
                pass

            def instance_method(self):
                pass

        class Child(Base):
            pass

        instance = Base()
        base_attrs = list(vars(Base).keys())
        instance_attrs = list(vars(instance).keys())
        child_attrs = list(vars(Child).keys())
        flexmock(Base).should_receive("class_method").once()
        flexmock(Base).should_receive("static_method").once()
        Base.class_method()
        Base.static_method()

        flexmock(instance).should_receive("class_method").once()
        flexmock(instance).should_receive("static_method").once()
        flexmock(instance).should_receive("instance_method").once()
        instance.class_method()
        instance.static_method()
        instance.instance_method()

        flexmock(Child).should_receive("class_method").once()
        flexmock(Child).should_receive("static_method").once()
        Child.class_method()
        Child.static_method()

        flexmock_teardown()
        assert base_attrs == list(vars(Base).keys())
        assert instance_attrs == list(vars(instance).keys())
        assert child_attrs == list(vars(Child).keys())

    def test_class_attributes_are_unchanged_after_spying(self):
        class Base:
            @classmethod
            def class_method(cls):
                pass

            @staticmethod
            def static_method():
                pass

            def instance_method(self):
                pass

        class Child(Base):
            pass

        instance = Base()
        base_attrs = list(vars(Base).keys())
        instance_attrs = list(vars(instance).keys())
        child_attrs = list(vars(Child).keys())
        flexmock(Base).should_call("class_method").times(3)  # TODO: should be once #80
        flexmock(Base).should_call("static_method").times(3)  # TODO: should be once #80
        Base.class_method()
        Base.static_method()

        flexmock(instance).should_call("class_method").once()
        flexmock(instance).should_call("static_method").once()
        flexmock(instance).should_call("instance_method").once()
        instance.class_method()
        instance.static_method()
        instance.instance_method()

        flexmock(Child).should_call("class_method").once()
        flexmock(Child).should_call("static_method").once()
        Child.class_method()
        Child.static_method()

        flexmock_teardown()
        assert base_attrs == list(vars(Base).keys())
        assert instance_attrs == list(vars(instance).keys())
        assert child_attrs == list(vars(Child).keys())

    def test_flexmock_should_cleanup_after_exception(self):
        class User:
            def method2(self):
                pass

        class Group:
            def method1(self):
                pass

        flexmock(Group)
        flexmock(User)
        Group.should_receive("method1").once()
        User.should_receive("method2").once()
        with assert_raises(
            exceptions.MethodCallError,
            "method1() expected to be called exactly 1 time, called 0 times",
        ):
            flexmock_teardown()
        for method in UPDATED_ATTRS:
            assert method not in dir(Group)
        for method in UPDATED_ATTRS:
            assert method not in dir(User)

    def test_flexmock_teardown_called_between_tests_part1(self):
        flexmock().should_receive("method1").ordered()

    def test_flexmock_teardown_called_between_tests_part2(self):
        mock = flexmock().should_receive("method2").ordered().mock()
        # Raises CallOrderError if flexmock teardown is not automatically called
        # after test part 1 above
        mock.method2()



"""Utility functions for testing."""

from contextlib import contextmanager
from typing import Type, Union

from flexmock._api import RE_TYPE


@contextmanager
def assert_raises(expected_exception: Type[BaseException], match: Union[RE_TYPE, str, None]):
    """Assert that code raises the correct exception with a correct error message.

    Args:
        expected_exception: Type of the expected exception.
        match: String or pattern to match the error message against. Use None
          to skip error message checking.
    """
    try:
        yield
    except Exception as raised_exception:
        if not isinstance(raised_exception, expected_exception):
            raise AssertionError(
                f"Expected exception '{type(expected_exception)}' "
                f"but '{type(raised_exception)}' was raised"
            ) from raised_exception
        if match is not None:
            fail = False
            if isinstance(match, RE_TYPE):
                fail = not match.search(str(raised_exception))
                match = match.pattern
            else:
                fail = str(raised_exception) != str(match)
            if fail:
                raise AssertionError(
                    f"Expected error message:\n{'-'*39}\n'{str(match)}'\n"
                    f"\nBut got:\n\n'{str(raised_exception)}'\n{'-'*39}\n"
                ) from raised_exception
    else:
        raise AssertionError(f"Exception '{expected_exception.__name__}' not raised")



"""Flexmock Sphinx documentation configuration file.

This file adds Sphinx support for documentation so that man pages can be
generated using Sphinx.
"""
import os
import sys

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.abspath(".."))

# -- Project information -----------------------------------------------------

project = "flexmock"
copyright = "2021, Slavek Kabrda, Herman Sheremetyev"

# -- General configuration ---------------------------------------------------

extensions = ["myst_parser"]
source_suffix = {
    ".rst": "restructuredtext",
    ".md": "markdown",
}
master_doc = "toctree"
exclude_patterns = ["api.md", "changelog.md", "contributing.md"]

# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ("toctree", "flexmock", "flexmock Documentation", ["Slavek Kabrda, Herman Sheremetyev"], 1)
]



"""Test unittest integration."""

# pylint: disable=missing-docstring
import sys
import unittest

from flexmock import flexmock
from tests.features import FlexmockTestCase


class TestUnitTestIntegration(FlexmockTestCase, unittest.TestCase):
    """Flexmock unittest integration specific tests."""

    def test_failed_test_case(self):
        """This tests that after a successful tests, failing flexmock assertions
        will change the test result from successful to failed.
        """
        flexmock().should_receive("this_test_should_fail").once()


if __name__ == "__main__":
    EXPECTED_FAILURES = 1
    test = unittest.main(exit=False)

    if (
        len(test.result.failures) == EXPECTED_FAILURES
        # Make sure that inherited tests are executed
        and test.result.testsRun > EXPECTED_FAILURES
    ):
        sys.exit(0)  # OK
    sys.exit(1)



"""Test testtools integration."""

# pylint: disable=missing-docstring
import testtools

from tests.features import FlexmockTestCase


class TestTestoolsIntegration(FlexmockTestCase, testtools.TestCase):
    pass



"""Test doctest integration and flexmock API examples."""

# pylint: disable=missing-docstring
import doctest
import sys

from flexmock import _api


class Plane:
    model = "Airpy 1"
    is_flying = True

    def __init__(self):
        self.speed = 3
        self.direction = None
        self.destination = None
        self.passengers = ["user1", "user2", "user3"]

    def pilot(self):
        return self.passengers[0]

    def fly(self, direction="up", destination="Helsinki"):
        self.direction = direction
        self.destination = destination

    def land(self):
        return None

    def set_speed(self, speed):
        self.speed = speed

    def repair(self, part):
        del part

    def crash(self):
        return None

    def flight_log(self):
        yield from ["land", "repair"]

    def passenger_count(self):
        return len(self.passengers)


class BadWeatherException(Exception):
    pass


class TestDoctestTeardown:
    """Test doctest runner teardown."""

    def test_flexmock_teardown_works_with_doctest_part_1(self):
        """Part 1

        Examples:
            Part 1:

            >>> from flexmock import flexmock
            >>> flexmock().should_receive("method1").ordered()
            <flexmock._api.Expectation object at ...>
        """

    def test_flexmock_teardown_works_with_doctest_part_2(self):
        """Raises CallOrderError if flexmock teardown is not automatically called
        after test part 1 above.

        Examples:
            Part 2:

            >>> from flexmock import flexmock
            >>> mock = flexmock().should_receive("method2").ordered().mock()
            >>> mock.method2()
        """


if __name__ == "__main__":
    results1 = doctest.testmod(
        sys.modules[__name__],  # current module
        optionflags=doctest.ELLIPSIS,
    )

    results2 = doctest.testmod(
        _api,
        extraglobs={
            "Plane": Plane,
            "plane": Plane(),
            "BadWeatherException": BadWeatherException,
        },
        optionflags=doctest.ELLIPSIS,
    )

    sys.exit(bool(results1.failed + results2.failed))



"""Tests for flexmock.

Tests under this module should not depend on any test runner.
"""

from .arg_matching import ArgumentMatchingTestCase
from .call_count import CallCountTestCase
from .common import CommonTestCase
from .conditional import ConditionalAssertionsTestCase
from .derived import DerivedTestCase
from .mocking import MockingTestCase
from .ordered import OrderedTestCase
from .proxied import ProxiedTestCase
from .spying import SpyingTestCase
from .stubbing import StubbingTestCase
from .teardown import TeardownTestCase


class FlexmockTestCase(
    ArgumentMatchingTestCase,
    CallCountTestCase,
    CommonTestCase,
    ConditionalAssertionsTestCase,
    DerivedTestCase,
    MockingTestCase,
    OrderedTestCase,
    ProxiedTestCase,
    SpyingTestCase,
    StubbingTestCase,
    TeardownTestCase,
):
    """This class should be imported by other test modules to run full flexmock
    test suite.
    """



"""Tests for argument matching."""

# pylint: disable=missing-docstring,no-member
import re

from flexmock import exceptions, flexmock
from flexmock._api import flexmock_teardown
from tests import some_module
from tests.some_module import SomeClass
from tests.utils import assert_raises


class ArgumentMatchingTestCase:
    def test_arg_matching_works_with_regexp(self):
        class Foo:
            def method(self, arg1, arg2):
                pass

        instance = Foo()
        flexmock(instance).should_receive("method").with_args(
            re.compile("^arg1.*asdf$"), arg2=re.compile("f")
        ).and_return("mocked")
        assert instance.method("arg1somejunkasdf", arg2="aadsfdas") == "mocked"

    def test_arg_matching_with_regexp_fails_when_regexp_doesnt_match_karg(self):
        class Foo:
            def method(self, arg1, arg2):
                pass

        instance = Foo()
        flexmock(instance).should_receive("method").with_args(
            re.compile("^arg1.*asdf$"), arg2=re.compile("a")
        ).and_return("mocked")
        with assert_raises(
            exceptions.MethodSignatureError,
            (
                "Arguments for call method did not match expectations:\n"
                '  Received call:\tmethod("arg1somejunkasdfa", arg2="a")\n'
                "  Expected call[1]:\tmethod(arg2=/a/, arg1=/^arg1.*asdf$/)"
            ),
        ):
            instance.method("arg1somejunkasdfa", arg2="a")

    def test_arg_matching_with_regexp_fails_when_regexp_doesnt_match_kwarg(self):
        class Foo:
            def method(self, arg1, arg2):
                pass

        instance = Foo()
        flexmock(instance).should_receive("method").with_args(
            re.compile("^arg1.*asdf$"), arg2=re.compile("a")
        ).and_return("mocked")
        with assert_raises(
            exceptions.MethodSignatureError,
            (
                "Arguments for call method did not match expectations:\n"
                '  Received call:\tmethod("arg1somejunkasdf", arg2="b")\n'
                "  Expected call[1]:\tmethod(arg2=/a/, arg1=/^arg1.*asdf$/)"
            ),
        ):
            instance.method("arg1somejunkasdf", arg2="b")

    def test_module_level_function_with_kwargs(self):
        flexmock(some_module).should_receive("module_function").with_args(1, y="expected")
        with assert_raises(
            exceptions.FlexmockError,
            (
                "Arguments for call module_function did not match expectations:\n"
                '  Received call:\tmodule_function(1, y="not expected")\n'
                '  Expected call[1]:\tmodule_function(y="expected", x=1)'
            ),
        ):
            some_module.module_function(1, y="not expected")

    def test_flexmock_should_match_types_on_multiple_arguments(self):
        class Foo:
            def method(self, arg1, arg2):
                pass

        instance = Foo()
        flexmock(instance).should_receive("method").with_args(str, int).and_return("ok")
        assert instance.method("some string", 12) == "ok"
        with assert_raises(
            exceptions.MethodSignatureError,
            (
                "Arguments for call method did not match expectations:\n"
                "  Received call:\tmethod(12, 32)\n"
                "  Expected call[1]:\tmethod(arg1=<class 'str'>, arg2=<class 'int'>)"
            ),
        ):
            instance.method(12, 32)
        flexmock(instance).should_receive("method").with_args(str, int).and_return("ok")
        with assert_raises(
            exceptions.MethodSignatureError,
            (
                "Arguments for call method did not match expectations:\n"
                '  Received call:\tmethod(12, "some string")\n'
                "  Expected call[1]:\tmethod(arg1=<class 'str'>, arg2=<class 'int'>)\n"
                "  Expected call[2]:\tmethod(arg1=<class 'str'>, arg2=<class 'int'>)"
            ),
        ):
            instance.method(12, "some string")
        flexmock(instance).should_receive("method").with_args(str, int).and_return("ok")
        with assert_raises(
            exceptions.MethodSignatureError,
            (
                "Arguments for call method did not match expectations:\n"
                '  Received call:\tmethod("string", 12, 14)\n'
                "  Expected call[1]:\tmethod(arg1=<class 'str'>, arg2=<class 'int'>)\n"
                "  Expected call[2]:\tmethod(arg1=<class 'str'>, arg2=<class 'int'>)\n"
                "  Expected call[3]:\tmethod(arg1=<class 'str'>, arg2=<class 'int'>)"
            ),
        ):
            instance.method("string", 12, 14)

    def test_flexmock_should_match_types_on_multiple_arguments_generic(self):
        class Foo:
            def method(self, a, b, c):  # pylint: disable=invalid-name
                pass

        instance = Foo()
        flexmock(instance).should_receive("method").with_args(object, object, object).and_return(
            "ok"
        )
        assert instance.method("some string", None, 12) == "ok"
        assert instance.method((1,), None, 12) == "ok"
        assert instance.method(12, 14, []) == "ok"
        assert instance.method("some string", "another one", False) == "ok"
        with assert_raises(
            exceptions.MethodSignatureError,
            (
                "Arguments for call method did not match expectations:\n"
                '  Received call:\tmethod("string", 12)\n'
                "  Expected call[1]:\tmethod(a=<class 'object'>, "
                "b=<class 'object'>, c=<class 'object'>)"
            ),
        ):
            instance.method("string", 12)  # pylint: disable=no-value-for-parameter
        flexmock_teardown()
        flexmock(instance).should_receive("method").with_args(object, object, object).and_return(
            "ok"
        )
        with assert_raises(
            exceptions.MethodSignatureError,
            (
                "Arguments for call method did not match expectations:\n"
                '  Received call:\tmethod("string", 12, 13, 14)\n'
                "  Expected call[1]:\tmethod(a=<class 'object'>, "
                "b=<class 'object'>, c=<class 'object'>)"
            ),
        ):
            instance.method("string", 12, 13, 14)

    def test_flexmock_should_match_types_on_multiple_arguments_classes(self):
        class Foo:
            def method(self, a, b):  # pylint: disable=invalid-name
                pass

        class Bar:
            pass

        foo_instance = Foo()
        bar_instance = Bar()
        flexmock(foo_instance).should_receive("method").with_args(object, Bar).and_return("ok")
        assert foo_instance.method("some string", bar_instance) == "ok"
        with assert_raises(
            exceptions.MethodSignatureError,
            re.compile(
                "Arguments for call method did not match expectations:\n"
                r'  Received call:\tmethod\(.+\.<locals>\.Bar object at 0x.+>, "some string"\)\n'
                r"  Expected call\[1\]:\tmethod\(a=<class 'object'>, b=<class.+\.<locals>\.Bar'>\)"
            ),
        ):
            foo_instance.method(bar_instance, "some string")
        flexmock_teardown()
        flexmock(foo_instance).should_receive("method").with_args(object, Bar).and_return("ok")
        with assert_raises(
            exceptions.MethodSignatureError,
            re.compile(
                "Arguments for call method did not match expectations:\n"
                r'  Received call:\tmethod\(12, "some string"\)\n'
                r"  Expected call\[1\]:\tmethod\(a=<class 'object'>, b=<class.+\.<locals>\.Bar'>\)"
            ),
        ):
            foo_instance.method(12, "some string")

    def test_flexmock_should_match_keyword_arguments(self):
        class Foo:
            def method(self, arg1, **kwargs):
                pass

        instance = Foo()
        flexmock(instance).should_receive("method").with_args(1, arg3=3, arg2=2).twice()
        instance.method(1, arg2=2, arg3=3)
        instance.method(1, arg3=3, arg2=2)
        flexmock_teardown()
        flexmock(instance).should_receive("method").with_args(1, arg3=3, arg2=2)
        with assert_raises(
            exceptions.MethodSignatureError,
            (
                "Arguments for call method did not match expectations:\n"
                "  Received call:\tmethod(arg2=2, arg3=3)\n"
                "  Expected call[1]:\tmethod(arg3=3, arg2=2, arg1=1)"
            ),
        ):
            instance.method(arg2=2, arg3=3)  # pylint: disable=no-value-for-parameter
        flexmock_teardown()
        flexmock(instance).should_receive("method").with_args(1, arg3=3, arg2=2)
        with assert_raises(
            exceptions.MethodSignatureError,
            (
                "Arguments for call method did not match expectations:\n"
                "  Received call:\tmethod(1, arg2=2, arg3=4)\n"
                "  Expected call[1]:\tmethod(arg3=3, arg2=2, arg1=1)"
            ),
        ):
            instance.method(1, arg2=2, arg3=4)
        flexmock_teardown()
        flexmock(instance).should_receive("method").with_args(1, arg3=3, arg2=2)
        with assert_raises(
            exceptions.MethodSignatureError,
            (
                "Arguments for call method did not match expectations:\n"
                "  Received call:\tmethod(1)\n"
                "  Expected call[1]:\tmethod(arg3=3, arg2=2, arg1=1)"
            ),
        ):
            instance.method(1)

    def test_flexmock_should_call_should_match_keyword_arguments(self):
        class Foo:
            def method(self, arg1, arg2=None, arg3=None):
                return f"{arg1}{arg2}{arg3}"

        instance = Foo()
        flexmock(instance).should_call("method").with_args(1, arg3=3, arg2=2).once()
        assert instance.method(1, arg2=2, arg3=3) == "123"

    def test_with_args_with_instance_method(self):
        flexmock(SomeClass).should_receive("instance_method_with_args").with_args("red").once()
        flexmock(SomeClass).should_receive("instance_method_with_args").with_args("blue").once()
        instance = SomeClass()
        instance.instance_method_with_args("red")
        instance.instance_method_with_args("blue")

    def test_with_args_with_class_method(self):
        flexmock(SomeClass).should_receive("class_method_with_args").with_args("red").once()
        flexmock(SomeClass).should_receive("class_method_with_args").with_args("blue").once()
        SomeClass.class_method_with_args("red")
        SomeClass.class_method_with_args("blue")

    def test_with_args_with_static_method(self):
        flexmock(SomeClass).should_receive("static_method_with_args").with_args("red").once()
        flexmock(SomeClass).should_receive("static_method_with_args").with_args("blue").once()
        SomeClass.static_method_with_args("red")
        SomeClass.static_method_with_args("blue")



"""Flexmock exceptions."""


class FlexmockError(Exception):
    """FlexmockError"""


class MockBuiltinError(Exception):
    """MockBuiltinError"""


class MethodSignatureError(FlexmockError):
    """MethodSignatureError"""


class ExceptionClassError(FlexmockError):
    """ExceptionClassError"""


class ExceptionMessageError(FlexmockError):
    """ExceptionMessageError"""


class StateError(FlexmockError):
    """StateError"""


class MethodCallError(FlexmockError):
    """MethodCallError"""


class CallOrderError(FlexmockError):
    """CallOrderError"""



"""Tests for mocking and spying derived classes."""

# pylint: disable=missing-docstring,no-member
from flexmock import exceptions, flexmock
from flexmock._api import flexmock_teardown
from tests.some_module import DerivedClass, SomeClass
from tests.utils import assert_raises


class DerivedTestCase:
    def test_mock_class_method_on_derived_class(self):
        flexmock(DerivedClass).should_receive("class_method").and_return(2).twice()
        assert DerivedClass().class_method() == 2
        assert DerivedClass.class_method() == 2

    def test_mock_class_method_on_derived_class_after_mocking_base_class(self):
        flexmock(SomeClass).should_receive("class_method").and_return(1).once()
        assert SomeClass.class_method() == 1
        flexmock(DerivedClass).should_receive("class_method").and_return(2).twice()
        assert DerivedClass().class_method() == 2
        assert DerivedClass.class_method() == 2

    def test_mock_static_method_on_derived_class(self):
        flexmock(DerivedClass).should_receive("static_method").and_return(4).twice()
        assert DerivedClass().static_method() == 4
        assert DerivedClass.static_method() == 4

    def test_mock_static_method_on_derived_class_after_mocking_base_class(self):
        flexmock(SomeClass).should_receive("static_method").and_return(3).once()
        assert SomeClass.static_method() == 3
        flexmock(DerivedClass).should_receive("static_method").and_return(4).twice()
        assert DerivedClass().static_method() == 4
        assert DerivedClass.static_method() == 4

    def test_mock_class_method_with_args_on_derived_class(self):
        flexmock(DerivedClass).should_receive("class_method_with_args").with_args(2).and_return(
            3
        ).twice()
        assert DerivedClass().class_method_with_args(2) == 3
        assert DerivedClass.class_method_with_args(2) == 3

    def test_mock_class_method_with_args_on_derived_class_after_mocking_base_class(self):
        flexmock(SomeClass).should_receive("class_method_with_args").with_args(1).and_return(
            2
        ).once()
        assert SomeClass.class_method_with_args(1) == 2
        flexmock(DerivedClass).should_receive("class_method_with_args").with_args(2).and_return(
            3
        ).twice()
        assert DerivedClass().class_method_with_args(2) == 3
        assert DerivedClass.class_method_with_args(2) == 3

    def test_mock_static_method_with_args_on_derived_class(self):
        flexmock(DerivedClass).should_receive("static_method_with_args").with_args(4).and_return(
            5
        ).twice()
        assert DerivedClass().static_method_with_args(4) == 5
        assert DerivedClass.static_method_with_args(4) == 5

    def test_mock_static_method_with_args_on_derived_class_after_mocking_base_class(self):
        flexmock(SomeClass).should_receive("static_method_with_args").with_args(2).and_return(
            3
        ).once()
        assert SomeClass.static_method_with_args(2) == 3
        flexmock(DerivedClass).should_receive("static_method_with_args").with_args(4).and_return(
            5
        ).twice()
        assert DerivedClass().static_method_with_args(4) == 5
        assert DerivedClass.static_method_with_args(4) == 5

    def test_spy_class_method_on_derived_class(self):
        flexmock(DerivedClass).should_call("class_method").and_return("class_method").twice()
        assert DerivedClass().class_method() == "class_method"
        assert DerivedClass.class_method() == "class_method"

    def test_spy_class_method_on_derived_class_after_spying_base_class(self):
        flexmock(SomeClass).should_call("class_method").and_return("class_method").times(
            3
        )  # TODO: Should be once #80
        assert SomeClass.class_method() == "class_method"
        flexmock(DerivedClass).should_call("class_method").and_return("class_method").twice()
        assert DerivedClass().class_method() == "class_method"
        assert DerivedClass.class_method() == "class_method"

    def test_spy_static_method_on_derived_class(self):
        flexmock(DerivedClass).should_call("static_method").and_return("static_method").twice()
        assert DerivedClass().static_method() == "static_method"
        assert DerivedClass.static_method() == "static_method"

    def test_spy_static_method_on_derived_class_after_spying_base_class(self):
        flexmock(SomeClass).should_call("static_method").and_return("static_method").times(
            3
        )  # TODO: Should be once #80
        assert SomeClass.static_method() == "static_method"
        flexmock(DerivedClass).should_call("static_method").and_return("static_method").twice()
        assert DerivedClass().static_method() == "static_method"
        assert DerivedClass.static_method() == "static_method"

    def test_spy_class_method_with_args_on_derived_class(self):
        flexmock(DerivedClass).should_call("class_method_with_args").with_args(2).and_return(2)
        assert DerivedClass().class_method_with_args(2) == 2
        assert DerivedClass.class_method_with_args(2) == 2

    @assert_raises(
        exceptions.MethodSignatureError, match=None
    )  # TODO: Should not raise exception #79
    def test_spy_class_method_with_args_on_derived_class_after_spying_base_class(self):
        flexmock(SomeClass).should_call("class_method_with_args").with_args(1).and_return(1)
        assert SomeClass.class_method_with_args(1) == 1
        flexmock(DerivedClass).should_call("class_method_with_args").with_args(2).and_return(2)
        assert DerivedClass().class_method_with_args(2) == 2
        assert DerivedClass.class_method_with_args(2) == 2

    def test_spy_static_method_with_args_on_derived_class(self):
        flexmock(DerivedClass).should_call("static_method_with_args").with_args(4).and_return(
            4
        ).twice()
        assert DerivedClass().static_method_with_args(4) == 4
        assert DerivedClass.static_method_with_args(4) == 4

    @assert_raises(
        exceptions.MethodSignatureError, match=None
    )  # TODO: Should not raise exception #79
    def test_spy_static_method_with_args_on_derived_class_after_spying_base_class(self):
        flexmock(SomeClass).should_call("static_method_with_args").with_args(2).and_return(2).once()
        assert SomeClass.static_method_with_args(2) == 2
        flexmock(DerivedClass).should_call("static_method_with_args").with_args(4).and_return(
            4
        ).once()  # should be twice
        assert DerivedClass().static_method_with_args(4) == 4
        assert DerivedClass.static_method_with_args(4) == 4

    def test_multi_mock_class_method_on_derived_class_is_properly_torn_down(self):
        flexmock(DerivedClass).should_receive("class_method_with_args").with_args(1).once()
        DerivedClass.class_method_with_args(1)

        flexmock(DerivedClass).should_receive("class_method_with_args").with_args(2).once()
        DerivedClass.class_method_with_args(2)

        flexmock(DerivedClass).should_receive("class_method_with_args").with_args(3).once()
        DerivedClass.class_method_with_args(3)

        # Run teardown manually to allow this test case to be contained in single test
        flexmock_teardown()

        flexmock(DerivedClass).should_receive("class_method_with_args").with_args(4).once()
        DerivedClass.class_method_with_args(4)



"""Tests for mocking and spying proxied objects."""

# pylint: disable=missing-docstring,no-member
from flexmock import flexmock
from tests import some_module
from tests.proxy import Proxy
from tests.some_module import DerivedClass, SomeClass


class ProxiedTestCase:
    def test_mock_proxied_derived_class_with_args(self):
        # pylint: disable=not-callable,invalid-name
        DerivedClassProxy = Proxy(DerivedClass)
        flexmock(DerivedClassProxy).should_receive("class_method_with_args").with_args(
            "a"
        ).and_return(2).twice()
        assert DerivedClassProxy().class_method_with_args("a") == 2
        assert DerivedClassProxy.class_method_with_args("a") == 2
        flexmock(DerivedClassProxy).should_receive("static_method_with_args").with_args(
            "b"
        ).and_return(3).twice()
        assert DerivedClassProxy().static_method_with_args("b") == 3
        assert DerivedClassProxy.static_method_with_args("b") == 3
        instance = DerivedClassProxy()
        flexmock(instance).should_receive("instance_method_with_args").with_args("c").and_return(
            4
        ).once()
        assert instance.instance_method_with_args("c") == 4

    def test_mock_proxied_module_function(self):
        # pylint: disable=not-callable
        some_module_proxy = Proxy(some_module)
        flexmock(some_module_proxy).should_receive("module_function").and_return(3).once()
        assert some_module_proxy.module_function() == 3

    def test_mock_proxied_derived_class(self):
        # pylint: disable=not-callable,invalid-name
        DerivedClassProxy = Proxy(DerivedClass)
        flexmock(DerivedClassProxy).should_receive("class_method").and_return(2).twice()
        assert DerivedClassProxy().class_method() == 2
        assert DerivedClassProxy.class_method() == 2
        flexmock(DerivedClassProxy).should_receive("static_method").and_return(3).twice()
        assert DerivedClassProxy().static_method() == 3
        assert DerivedClassProxy.static_method() == 3
        instance = DerivedClassProxy()
        flexmock(instance).should_receive("instance_method").and_return(4).once()
        assert instance.instance_method() == 4

    def test_mock_proxied_class_with_args(self):
        # pylint: disable=not-callable,invalid-name
        SomeClassProxy = Proxy(SomeClass)
        flexmock(SomeClassProxy).should_receive("class_method_with_args").with_args("a").and_return(
            2
        ).twice()
        assert SomeClassProxy().class_method_with_args("a") == 2
        assert SomeClassProxy.class_method_with_args("a") == 2
        flexmock(SomeClassProxy).should_receive("static_method_with_args").with_args(
            "b"
        ).and_return(3).twice()
        assert SomeClassProxy().static_method_with_args("b") == 3
        assert SomeClassProxy.static_method_with_args("b") == 3
        instance = SomeClassProxy()
        flexmock(instance).should_receive("instance_method_with_args").with_args("c").and_return(
            4
        ).once()
        assert instance.instance_method_with_args("c") == 4

    def test_mock_proxied_class(self):
        # pylint: disable=not-callable,invalid-name
        SomeClassProxy = Proxy(SomeClass)
        flexmock(SomeClassProxy).should_receive("class_method").and_return(2).twice()
        assert SomeClassProxy().class_method() == 2
        assert SomeClassProxy.class_method() == 2
        flexmock(SomeClassProxy).should_receive("static_method").and_return(3).twice()
        assert SomeClassProxy().static_method() == 3
        assert SomeClassProxy.static_method() == 3
        instance = SomeClassProxy()
        flexmock(instance).should_receive("instance_method").and_return(4).once()
        assert instance.instance_method() == 4

    def test_spy_proxied_derived_class(self):
        # pylint: disable=not-callable,invalid-name
        DerivedClassProxy = Proxy(DerivedClass)
        flexmock(DerivedClassProxy).should_call("class_method").and_return("class_method").twice()
        assert DerivedClassProxy().class_method() == "class_method"
        assert DerivedClassProxy.class_method() == "class_method"
        flexmock(DerivedClassProxy).should_call("static_method").and_return("static_method").twice()
        assert DerivedClassProxy().static_method() == "static_method"
        assert DerivedClassProxy.static_method() == "static_method"
        instance = DerivedClassProxy()
        flexmock(instance).should_call("instance_method").and_return("instance_method").once()
        assert instance.instance_method() == "instance_method"

    def test_spy_proxied_derived_class_with_args(self):
        # pylint: disable=not-callable,invalid-name
        DerivedClassProxy = Proxy(DerivedClass)
        flexmock(DerivedClassProxy).should_call("class_method_with_args").with_args("a").and_return(
            "a"
        ).twice()
        assert DerivedClassProxy().class_method_with_args("a") == "a"
        assert DerivedClassProxy.class_method_with_args("a") == "a"
        flexmock(DerivedClassProxy).should_call("static_method_with_args").with_args(
            "b"
        ).and_return("b").twice()
        assert DerivedClassProxy().static_method_with_args("b") == "b"
        assert DerivedClassProxy.static_method_with_args("b") == "b"
        instance = DerivedClassProxy()
        flexmock(instance).should_call("instance_method_with_args").with_args("c").and_return(
            "c"
        ).once()
        assert instance.instance_method_with_args("c") == "c"

    def test_spy_proxied_module_function(self):
        # pylint: disable=not-callable
        some_module_proxy = Proxy(some_module)
        flexmock(some_module_proxy).should_receive("module_function").and_return(0).once()
        assert some_module_proxy.module_function(2, 2) == 0

    def test_spy_proxied_class(self):
        # pylint: disable=not-callable,invalid-name
        SomeClassProxy = Proxy(SomeClass)
        flexmock(SomeClassProxy).should_call("class_method").and_return("class_method").twice()
        assert SomeClassProxy().class_method() == "class_method"
        assert SomeClassProxy.class_method() == "class_method"
        flexmock(SomeClassProxy).should_call("static_method").and_return("static_method").twice()
        assert SomeClassProxy().static_method() == "static_method"
        assert SomeClassProxy.static_method() == "static_method"
        instance = SomeClassProxy()
        flexmock(instance).should_call("instance_method").and_return("instance_method").once()
        assert instance.instance_method() == "instance_method"

    def test_spy_proxied_class_with_args(self):
        # pylint: disable=not-callable,invalid-name
        SomeClassProxy = Proxy(SomeClass)
        flexmock(SomeClassProxy).should_call("class_method_with_args").with_args("a").and_return(
            "a"
        ).twice()
        assert SomeClassProxy().class_method_with_args("a") == "a"
        assert SomeClassProxy.class_method_with_args("a") == "a"
        flexmock(SomeClassProxy).should_call("static_method_with_args").with_args("b").and_return(
            "b"
        ).twice()
        assert SomeClassProxy().static_method_with_args("b") == "b"
        assert SomeClassProxy.static_method_with_args("b") == "b"
        instance = SomeClassProxy()
        flexmock(instance).should_call("instance_method_with_args").with_args("c").and_return(
            "c"
        ).once()
        assert instance.instance_method_with_args("c") == "c"



"""Tests for flexmock spying feature."""

# pylint: disable=missing-docstring,no-member,broad-exception-raised
import re

from flexmock import exceptions, flexmock
from flexmock._api import flexmock_teardown
from tests.utils import assert_raises


class SpyingTestCase:
    def test_spying_non_existent_mock_object_method_should_fail(self):
        mock = flexmock()
        with assert_raises(
            exceptions.FlexmockError,
            "Mock object does not have attribute 'method_foo'. "
            'Did you mean to call should_receive("method_foo") instead?',
        ):
            mock.should_call("method_foo")
        mock = flexmock(method_foo=lambda: "ok")
        mock.should_call("method_foo")

    def test_pass_thru_should_call_original_method_only_once(self):
        class Nyan:
            def __init__(self):
                self.value = 0

            def method(self):
                self.value += 1

        obj = Nyan()
        flexmock(obj)
        obj.should_call("method")
        obj.method()
        assert obj.value == 1

    def test_should_call_works_for_same_method_with_different_args(self):
        class Foo:
            def method(self, arg):
                pass

        instance = Foo()
        flexmock(instance).should_call("method").with_args("foo").once()
        flexmock(instance).should_call("method").with_args("bar").once()
        instance.method("foo")
        instance.method("bar")
        flexmock_teardown()

    def test_should_call_fails_properly_for_same_method_with_different_args(self):
        class Foo:
            def method(self, arg):
                pass

        instance = Foo()
        flexmock(instance).should_call("method").with_args("foo").once()
        flexmock(instance).should_call("method").with_args("bar").once()
        instance.method("foo")
        with assert_raises(
            exceptions.MethodCallError,
            'method(arg="bar") expected to be called exactly 1 time, called 0 times',
        ):
            flexmock_teardown()

    def test_should_call_alias_should_create_a_spy(self):
        class Foo:
            def get_stuff(self):
                return "yay"

        instance = Foo()
        flexmock(instance).should_call("get_stuff").and_return("yay").once()
        with assert_raises(
            exceptions.MethodCallError,
            "get_stuff() expected to be called exactly 1 time, called 0 times",
        ):
            flexmock_teardown()

    def test_spy_should_not_match_falsy_stuff(self):
        class Foo:
            def method1(self):
                return None

            def method2(self):
                return False

            def method3(self):
                return []

            def method4(self):
                return ""

        instance = Foo()
        flexmock(instance).should_call("method1").and_return("bar").once()
        flexmock(instance).should_call("method2").and_return("bar").once()
        flexmock(instance).should_call("method3").and_return("bar").once()
        flexmock(instance).should_call("method4").and_return("bar").once()
        with assert_raises(
            exceptions.FlexmockError,
            (
                "Returned values for call method1 did not match expectation:\n"
                "  Expected:\t'bar'\n"
                "  Returned:\tNone"
            ),
        ):
            instance.method1()
        with assert_raises(
            exceptions.FlexmockError,
            (
                "Returned values for call method2 did not match expectation:\n"
                "  Expected:\t'bar'\n"
                "  Returned:\tFalse"
            ),
        ):
            instance.method2()
        with assert_raises(
            exceptions.FlexmockError,
            (
                "Returned values for call method3 did not match expectation:\n"
                "  Expected:\t'bar'\n"
                "  Returned:\t[]"
            ),
        ):
            instance.method3()
        with assert_raises(
            exceptions.FlexmockError,
            (
                "Returned values for call method4 did not match expectation:\n"
                "  Expected:\t'bar'\n"
                "  Returned:\t''"
            ),
        ):
            instance.method4()

    def test_spy_should_match_return_value_class(self):
        class User:
            pass

        user = User()
        instance = flexmock(
            foo=lambda: ("bar", "baz"),
            bar=lambda: user,
            baz=lambda: None,
            bax=lambda: None,
        )
        instance.should_call("foo").and_return(str, str)
        instance.should_call("bar").and_return(User)
        instance.should_call("baz").and_return(object)
        instance.should_call("bax").and_return(None)
        assert instance.foo() == ("bar", "baz")
        assert instance.bar() == user
        assert instance.baz() is None
        assert instance.bax() is None

    def test_flexmock_spy_should_not_clobber_original_method(self):
        class User:
            def get_stuff(self):
                return "real", "stuff"

        user = User()
        flexmock(user).should_call("get_stuff")
        flexmock(user).should_call("get_stuff")
        assert user.get_stuff() == ("real", "stuff")

    def test_and_raise_with_value_that_is_not_a_class(self):
        class RaisesException:
            def get_stuff(self):
                raise RuntimeError("baz")

        instance = RaisesException()
        flexmock(instance).should_call("get_stuff").and_raise(RuntimeError("foo")).once()
        with assert_raises(
            exceptions.ExceptionClassError,
            re.compile(
                r"Raised exception for call get_stuff did not match expectation:\n"
                # Python 3.6 contains comma after 'foo'
                r"  Expected:\tRuntimeError\('foo',?\)\n"
                r"  Raised:\t<class 'RuntimeError'>\n\n"
                r"Did you try to call and_raise with an instance\?\n"
                r'Instead of and_raise\(Exception\("arg"\)\), try and_raise\(Exception, "arg"\)',
            ),
        ):
            instance.get_stuff()

    def test_flexmock_should_blow_up_on_wrong_spy_return_values(self):
        class User:
            def get_stuff(self):
                return "real", "stuff"

            def get_more_stuff(self):
                return "other", "stuff"

        user = User()
        flexmock(user).should_call("get_stuff").and_return("other", "stuff")
        with assert_raises(
            exceptions.MethodSignatureError,
            (
                "Returned values for call get_stuff did not match expectation:\n"
                "  Expected:\t('other', 'stuff')\n"
                "  Returned:\t('real', 'stuff')"
            ),
        ):
            user.get_stuff()
        flexmock(user).should_call("get_more_stuff").and_return()
        with assert_raises(
            exceptions.MethodSignatureError,
            (
                "Returned values for call get_more_stuff did not match expectation:\n"
                "  Expected:\tNone\n"
                "  Returned:\t('other', 'stuff')"
            ),
        ):
            user.get_more_stuff()

    def test_flexmock_should_reraise_exception_in_spy_with_return_values(self):
        class RaisesException:
            @classmethod
            def class_method(cls):
                raise RuntimeError("foo")

            @staticmethod
            def static_method():
                raise RuntimeError("bar")

            def instance_method(self):
                raise RuntimeError("baz")

        instance = RaisesException()
        flexmock(instance).should_call("class_method").and_return("foo").once()
        flexmock(instance).should_call("static_method").and_return("bar").once()
        flexmock(instance).should_call("instance_method").and_return("baz").once()

        with assert_raises(RuntimeError, "foo"):
            instance.class_method()
        with assert_raises(RuntimeError, "bar"):
            instance.static_method()
        with assert_raises(RuntimeError, "baz"):
            instance.instance_method()

        flexmock(RaisesException).should_call("class_method").once()
        flexmock(RaisesException).should_call("static_method").once()
        with assert_raises(RuntimeError, "foo"):
            RaisesException.class_method()
        with assert_raises(RuntimeError, "bar"):
            RaisesException.static_method()

    def test_flexmock_should_reraise_exception_in_spy(self):
        class RaisesException:
            @classmethod
            def class_method(cls):
                raise RuntimeError("foo")

            @staticmethod
            def static_method():
                raise RuntimeError("bar")

            def instance_method(self):
                raise RuntimeError("baz")

        instance = RaisesException()
        flexmock(instance).should_call("class_method").once()
        flexmock(instance).should_call("static_method").once()
        flexmock(instance).should_call("instance_method").once()

        with assert_raises(RuntimeError, "foo"):
            instance.class_method()
        with assert_raises(RuntimeError, "bar"):
            instance.static_method()
        with assert_raises(RuntimeError, "baz"):
            instance.instance_method()

        flexmock(RaisesException).should_call("class_method").once()
        flexmock(RaisesException).should_call("static_method").once()
        with assert_raises(RuntimeError, "foo"):
            RaisesException.class_method()
        with assert_raises(RuntimeError, "bar"):
            RaisesException.static_method()

    def test_flexmock_should_verify_correct_spy_return_values(self):
        class User:
            def get_stuff(self):
                return "real", "stuff"

        user = User()
        flexmock(user).should_call("get_stuff").and_return("real", "stuff")
        assert user.get_stuff() == ("real", "stuff")

    def test_flexmock_should_verify_correct_spy_regexp_return_values(self):
        class User:
            def get_stuff(self):
                return "real", "stuff"

        user = User()
        flexmock(user).should_call("get_stuff").and_return(
            re.compile("ea.*"), re.compile("^stuff$")
        )
        assert user.get_stuff() == ("real", "stuff")

    def test_flexmock_should_verify_spy_raises_correct_exception_class(self):
        class FakeException(Exception):
            def __init__(self, param, param2):
                self.message = f"{param}, {param2}"
                Exception.__init__(self)

        class User:
            def get_stuff(self):
                raise FakeException(1, 2)

        user = User()
        flexmock(user).should_call("get_stuff").and_raise(FakeException, 1, 2)
        user.get_stuff()

    def test_flexmock_should_verify_spy_matches_exception_message(self):
        class FakeException(Exception):
            def __init__(self, param1, param2):
                self.param1 = param1
                self.param2 = param2
                Exception.__init__(self, param1)

            def __str__(self):
                return f"{self.param1}, {self.param2}"

        class User:
            def get_stuff(self):
                raise FakeException("1", "2")

        user = User()
        flexmock(user).should_call("get_stuff").and_raise(FakeException, "2", "1")
        with assert_raises(
            exceptions.ExceptionMessageError,
            (
                "Error message mismatch with raised FakeException:\n"
                "  Expected message:\n\t'1, 2'\n"
                "  Received message:\n\t'2, 1'"
            ),
        ):
            user.get_stuff()

    def test_flexmock_should_verify_spy_matches_exception_regexp(self):
        class User:
            def get_stuff(self):
                raise Exception("123asdf345")

        user = User()
        flexmock(user).should_call("get_stuff").and_raise(Exception, re.compile("asdf"))
        user.get_stuff()
        flexmock_teardown()

    def test_flexmock_should_verify_spy_matches_exception_regexp_mismatch(self):
        class User:
            def get_stuff(self):
                raise Exception("123asdf345")

        user = User()
        flexmock(user).should_call("get_stuff").and_raise(Exception, re.compile("^asdf"))
        with assert_raises(
            exceptions.ExceptionMessageError,
            (
                "Error message mismatch with raised Exception:\n"
                "  Expected pattern:\n\t/^asdf/\n"
                "  Received message:\n\t'123asdf345'"
            ),
        ):
            user.get_stuff()

    def test_flexmock_should_blow_up_on_wrong_spy_exception_type(self):
        class User:
            def get_stuff(self):
                raise exceptions.CallOrderError("foo")

        user = User()
        flexmock(user).should_call("get_stuff").and_raise(exceptions.MethodCallError)
        with assert_raises(
            exceptions.ExceptionClassError,
            (
                "Raised exception for call get_stuff did not match expectation:\n"
                "  Expected:\t<class 'flexmock.exceptions.MethodCallError'>\n"
                "  Raised:\t<class 'flexmock.exceptions.CallOrderError'>"
            ),
        ):
            user.get_stuff()

    def test_flexmock_should_match_spy_exception_parent_type(self):
        class User:
            def get_stuff(self):
                raise exceptions.CallOrderError("foo")

        user = User()
        flexmock(user).should_call("get_stuff").and_raise(exceptions.FlexmockError)
        user.get_stuff()

    def test_flexmock_should_call_respects_matched_expectations(self):
        class Group:
            def method1(self, arg1, arg2="b"):
                return f"{arg1}:{arg2}"

            def method2(self, arg):
                return arg

        group = Group()
        flexmock(group).should_call("method1").twice()
        assert group.method1("a", arg2="c") == "a:c"
        assert group.method1("a") == "a:b"
        group.should_call("method2").once().with_args("c")
        assert group.method2("c") == "c"
        flexmock_teardown()

    def test_flexmock_should_call_respects_unmatched_expectations(self):
        class Group:
            def method1(self, arg1, arg2="b"):
                return f"{arg1}:{arg2}"

            def method2(self, arg1):
                pass

        group = Group()
        flexmock(group).should_call("method1").at_least().once()
        with assert_raises(
            exceptions.MethodCallError,
            "method1() expected to be called at least 1 time, called 0 times",
        ):
            flexmock_teardown()
        flexmock(group)
        group.should_call("method2").with_args("a").once()
        group.should_receive("method2").with_args("not a")
        group.method2("not a")
        with assert_raises(
            exceptions.MethodCallError,
            'method2(arg1="a") expected to be called exactly 1 time, called 0 times',
        ):
            flexmock_teardown()

    def test_flexmock_should_not_blow_up_on_should_call_for_class_methods(self):
        class User:
            @classmethod
            def method(cls):
                return "class"

        flexmock(User).should_call("method")
        assert User.method() == "class"

    def test_flexmock_should_not_blow_up_on_should_call_for_static_methods(self):
        class User:
            @staticmethod
            def method():
                return "static"

        flexmock(User).should_call("method")
        assert User.method() == "static"

    def test_should_call_with_class_default_attributes(self):
        """Flexmock should not allow mocking class default attributes like
        __call__ on an instance.
        """

        class WithCall:
            def __call__(self, arg1):
                return arg1

        instance = WithCall()

        with assert_raises(
            exceptions.FlexmockError,
            re.compile(r".+<locals>\.WithCall object at 0x.+> does not have attribute '__call__'"),
        ):
            flexmock(instance).should_call("__call__")

    def test_should_call_on_class_mock(self):
        class User:
            def __init__(self):
                self.value = "value"

            def method1(self):
                return "class"

            def method2(self):
                return self.value

        # Access class-level method
        user1 = User()
        user2 = User()
        flexmock(User).should_call("method1").once()
        with assert_raises(
            exceptions.MethodCallError,
            "method1() expected to be called exactly 1 time, called 0 times",
        ):
            flexmock_teardown()
        flexmock(User).should_call("method1").twice()
        assert user1.method1() == "class"
        assert user2.method1() == "class"

        # Access instance attributes
        flexmock(User).should_call("method2").once()
        with assert_raises(
            exceptions.MethodCallError,
            "method2() expected to be called exactly 1 time, called 0 times",
        ):
            flexmock_teardown()
        flexmock(User).should_call("method2").twice()
        assert user1.method2() == "value"
        assert user2.method2() == "value"

        # Try resetting the expectation
        flexmock(User).should_call("method2").once()
        assert user1.method2() == "value"



"""Hook flexmock into Pytest.
Pytest is a Python test framework:
https://github.com/pytest-dev/pytest
"""

from typing import Generator, Optional

import pytest
from _pytest.runner import CallInfo, ExceptionInfo, Item, TestReport

from flexmock._api import flexmock_teardown


@pytest.hookimpl(hookwrapper=True)
def pytest_runtest_makereport(
    item: Item,  # pylint: disable=unused-argument
    call: CallInfo[None],
) -> Generator[None, None, None]:
    """Called to verify and tear down the mocks set up by flexmock.

    Args:
        item: Test item for which the hook is being called.
        call: Information about the current call phase.

    Yields:
        Control back to the pytest runner.
    """
    if call.when == "call":
        try:
            flexmock_teardown()
        except BaseException:  # pylint: disable=broad-except
            call.excinfo = call.excinfo or ExceptionInfo.from_current()
    elif call.when == "teardown":
        flexmock_teardown()

    _test_report: Optional[TestReport] = yield



"""Flexmock test runner integrations."""

# pylint: disable=import-outside-toplevel, import-error
import sys
import unittest
from contextlib import suppress
from functools import wraps
from typing import Any, Type

from flexmock._api import flexmock_teardown


def _patch_test_result(klass: Type[Any]) -> None:
    """Patches flexmock into any class that inherits unittest.TestResult.

    This seems to work well for majority of test runners. In the case of nose
    it's not even necessary as it doesn't override unittest.TestResults's
    addSuccess and addFailure methods so simply patching unittest works
    out of the box for nose.

    For those that do inherit from unittest.TestResult and override its
    stopTest and addSuccess methods, patching is pretty straightforward
    (numerous examples below).

    The reason we don't simply patch unittest's parent TestResult class
    is stopTest and addSuccess in the child classes tend to add messages
    into the output that we want to override in case flexmock generates
    its own failures.

    Args:
        klass: the class whose stopTest method needs to be decorated.
    """
    # pylint: disable=invalid-name
    _patch_stop_test(klass)
    _patch_add_success(klass)


def _patch_stop_test(klass: Type[unittest.TextTestResult]) -> None:
    """Patch TextTestResult class stopTest method and add call to flexmock
    teardown.

    If the test failed already before flexmock teardown, nothing is done.
    However if the test was successful before flexmock teardown and flexmock
    assertions fail, flexmock updates the test result to failed by calling
    addFailure method.

    Args:
        klass: the class whose stopTest method needs to be decorated.
    """
    saved_add_success = klass.addSuccess
    saved_stop_test = klass.stopTest

    @wraps(saved_stop_test)
    def decorated(self: unittest.TextTestResult, test: unittest.TestCase) -> None:
        if saved_stop_test.__code__ is not decorated.__code__:  # type: ignore [attr-defined]
            # if parent class was for some reason patched, avoid calling
            # flexmock_teardown() twice and delegate up the class hierarchy
            # this doesn't help if there is a gap and only the parent's
            # parent class was patched, but should cover most screw-ups
            try:
                flexmock_teardown()
                saved_add_success(self, test)
            except Exception:
                if hasattr(self, "_pre_flexmock_success"):
                    self.addFailure(test, sys.exc_info())
            if hasattr(self, "_pre_flexmock_success"):
                del self._pre_flexmock_success
        return saved_stop_test(self, test)

    if klass.stopTest is not decorated:
        klass.stopTest = decorated  # type: ignore


def _patch_add_success(klass: Type[unittest.TextTestResult]) -> None:
    """Patch the addSuccess method of the TextTestResult class.

    TextTestResult addSuccess method is replaced and the original addSuccess
    method is called after flexmock teardown patched stopTest method.

    An attribute is set in the replaced addSuccess method to indicate if the
    test was successful before flexmock teardown was called.

    Args:
        klass: the class whose addSuccess method needs to be decorated.
    """

    @wraps(klass.addSuccess)
    def decorated(self: unittest.TextTestResult, _test: unittest.TestCase, **_kwargs: Any) -> None:
        self._pre_flexmock_success = True  # type: ignore

    if klass.addSuccess is not decorated:
        klass.addSuccess = decorated  # type: ignore


# Hook flexmock into doctest.

with suppress(ImportError):
    from doctest import DocTestRunner, TestResults

    saved_doctest = DocTestRunner.run

    @wraps(saved_doctest)
    def run(
        *args: Any,
        **kwargs: Any,
    ) -> TestResults:
        """Call the teardown at the end of the tests.

        Args:
            args: the arguments passed to the runner
            kwargs: the keyword arguments passed to the runner

        Returns:
            The test results

        """
        try:
            return saved_doctest(*args, **kwargs)
        finally:
            flexmock_teardown()

    DocTestRunner.run = run  # type: ignore


# Hook flexmock into unittest.
# only valid TestResult class for unittest is TextTestResult
_patch_test_result(unittest.TextTestResult)


# Hook into Teamcity unittests.
# This allows flexmock to be used within PyCharm.
with suppress(ImportError):
    from teamcity.unittestpy import TeamcityTestResult

    _patch_test_result(TeamcityTestResult)


# Hook into teststools.
# testtools is a set of extensions to the Python standard library's unit testing framework:
# https://github.com/testing-cabal/testtools

with suppress(ImportError):
    from testtools import testresult

    _patch_test_result(testresult.TestResult)


# Hook into Zope testrunner.
# Zope is an open-source web application server:
# https://github.com/zopefoundation/Zope

with suppress(ImportError):
    from zope import testrunner  # pylint: disable=no-name-in-module

    try:
        _patch_test_result(testrunner.runner.TestResult)
    except AttributeError:
        # testrunner.runner is only available when tests are executed with zope.testrunner
        pass


# Hook into subunit.
# Subunit is a test reporting and controlling protocol.
# https://github.com/testing-cabal/subunit

with suppress(ImportError):
    import subunit

    _patch_test_result(subunit.TestProtocolClient)
    with suppress(AttributeError):
        _patch_test_result(subunit.test_results.TestResultDecorator)


# Hook into twisted.
# Twisted is an event-based framework for internet applications:
# https://github.com/twisted/twisted

with suppress(ImportError):
    from twisted.trial import reporter

    _patch_test_result(reporter.MinimalReporter)
    _patch_test_result(reporter.TextReporter)
    _patch_test_result(reporter.VerboseTextReporter)
    _patch_test_result(reporter.TreeReporter)



"""Flexmock setup.py."""

from setuptools import setup

with open("README.md", encoding="utf-8") as file:
    long_description = file.read()

VERSION = "0.12.2"

setup(
    name="flexmock",
    version=VERSION,
    author="Slavek Kabrda, Herman Sheremetyev",
    author_email="slavek@redhat.com",
    url="https://flexmock.readthedocs.io/",
    project_urls={
        "Documentation": "https://flexmock.readthedocs.io/",
        "Changes": "https://flexmock.readthedocs.io/en/latest/changelog/",
        "Source Code": "https://github.com/flexmock/flexmock",
        "Issue Tracker": "https://github.com/flexmock/flexmock/issues",
    },
    license="BSD-2-Clause",
    description="flexmock is a testing library for Python that makes it easy to create mocks,"
    "stubs and fakes.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="mock testing test unittest pytest",
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Intended Audience :: Developers",
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Software Development :: Testing",
        "Topic :: Software Development :: Testing :: Mocking",
        "Topic :: Software Development :: Testing :: Unit",
        "Typing :: Typed",
    ],
    python_requires=">=3.8.0,<4.0.0",
    packages=["flexmock"],
    package_dir={"": "src"},
    include_package_data=True,
    command_options={
        "build_sphinx": {
            "version": ("setup.py", VERSION),
            "release": ("setup.py", VERSION),
        }
    },
    entry_points={
        'pytest11': [
            'flexmock = flexmock._pytest_plugin',
        ],
    },
)
